// This file is part of VKF, a CPython library
// for Machine Learning based on Lattice Theory.
//
// This Source Code Form is subject to the terms of the X11 (MIT) License.
// If a copy of the X11 License was not distributed with this file, 
// You can obtain one at https://opensource.org/licenses/MIT/.
//
// X11 (MIT) License
// 
// Copyright (c) 2020 Dmitry V. Vinogradov
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#include "Join.hpp"

double vkf::SemiRegress::Optimize(const vkf::Join &sample, const double threshold) {
    std::uint32_t sample_size = sample.size();
    double divergence = 0.5*sample_size;
    double old_div = 0.0;
    old_div = divergence + 2*threshold;
    while (old_div - divergence > threshold) {
        double m00 = 0.0;
        double m01 = 0.0;
        double m11 = 0.0;
        double r0 = 0.0;
        double r1 = 0.0;
        for (std::uint32_t ndex = 0; ndex < sample_size; ndex++) {
            double s;
            if (sample[ndex].IsPlus()) {
                s = 1.0/(1.0+std::exp(w0 + w1*sample[ndex][position]));
            } else {
                s = 1.0/(1.0+std::exp(-w0 - w1*sample[ndex][position]));
            }
            m00 += s*(1.0-s);
            m01 += sample[ndex][position]*s*(1.0-s);
            m11 += sample[ndex][position]*sample[ndex][position]*s*(1.0-s);
            if (sample[ndex].IsPlus()) {
                r0 += s;
                r1 += sample[ndex][position]*s;
            } else {
                r0 -= s;
                r1 -= sample[ndex][position]*s;
            }
        }
        double det = m00*m11 - m01*m01;
        if (det < 0.0001) { // ridge-regression
//            m00 += 0.05;
//            m11 += 0.05;
//            det = m00*m11 - m01*m01;
            break;
        }
        w0 += r0*m11/det - r1*m01/det;
        w1 += -r0*m01/det + r1*m00/det;
        old_div = divergence;
        divergence = 0.0;
        for(std::uint32_t ndex = 0; ndex < sample_size; ndex++) {
            if (sample[ndex].IsPlus()) {
                divergence += (1.0 - this->Predict(sample, ndex));
            } else {
                divergence += this->Predict(sample, ndex);
            }
        }
    }
    double loglikelyhood = 0.0;
//    double likelyhood = 1.0;
    for(std::uint32_t ndex = 0; ndex < sample_size; ndex++) {
        if (sample[ndex].IsPlus()) {
            loglikelyhood -=  std::log(1.0 + std::exp(-w0 - w1*sample[ndex][position]));
//            likelyhood *= 1.0/(1.0 + std::exp(-w0 - w1*sample[ndex][position]));
        } else {
            loglikelyhood -=  std::log(1.0 + std::exp(w0 + w1*sample[ndex][position]));
//            likelyhood *= 1.0/(1.0 + std::exp(w0 + w1*sample[ndex][position]));
        }
    }
    return loglikelyhood;
//    return likelyhood;
}


double vkf::Regress::Optimize(const vkf::Join &sample, const double threshold) {
    std::uint32_t sample_size = sample.size();
    double divergence = 0.5*sample_size;
    double old_div = 0.0;
    old_div = divergence + 2*threshold;
    while (old_div - divergence > threshold) {
        double m00 = 0.0;
        double m01 = 0.0;
        double m02 = 0.0;
        double m11 = 0.0;
        double m12 = 0.0;
        double m22 = 0.0;
        double r0 = 0.0;
        double r1 = 0.0;
        double r2 = 0.0;
        for (std::uint32_t ndex = 0; ndex < sample_size; ndex++) {
            double s;
            if (sample[ndex].IsPlus()) {
                s = 1.0/(1.0+std::exp(w0 + w1*sample[ndex][pos1] + w2*sample[ndex][pos2]));
            } else {
                s = 1.0/(1.0+std::exp(-w0 - w1*sample[ndex][pos1] - w2*sample[ndex][pos2]));
            }
            m00 += s*(1.0-s);
            m01 += sample[ndex][pos1]*s*(1.0-s);
            m02 += sample[ndex][pos2]*s*(1.0-s);
            m11 += sample[ndex][pos1]*sample[ndex][pos1]*s*(1.0-s);
            m12 += sample[ndex][pos1]*sample[ndex][pos2]*s*(1.0-s);
            m22 += sample[ndex][pos2]*sample[ndex][pos2]*s*(1.0-s);
            if (sample[ndex].IsPlus()) {
                r0 += s;
                r1 += sample[ndex][pos1]*s;
                r2 += sample[ndex][pos2]*s;
            } else {
                r0 -= s;
                r1 -= sample[ndex][pos1]*s;
                r2 -= sample[ndex][pos2]*s;
            }
        }
        double det = m00*m11*m22 + m01*m12*m02 + m02*m01*m12 - m02*m11*m02 - m12*m12*m00 - m22*m01*m01;
        if (det < 0.000001) { // ridge-regression
//            m00 += 0.05;
//            m11 += 0.05;
//            m22 += 0.05;
//            det = m00*m11*m22 + m01*m12*m02 + m02*m01*m12 - m02*m11*m02 - m12*m12*m00 - m22*m01*m01;
            break;
        }
        w0 += r0*(m11*m22-m12*m12)/det + r1*(m02*m12-m01*m22)/det + r2*(m01*m12-m02*m11)/det;
        w1 += r0*(m02*m12-m01*m22)/det + r1*(m00*m22-m02*m02)/det + r2*(m01*m02-m00*m12)/det;
        w2 += r0*(m01*m12-m02*m11)/det + r1*(m01*m02-m00*m12)/det + r2*(m00*m11-m01*m01)/det;
        old_div = divergence;
        divergence = 0.0;
        for(std::uint32_t ndex = 0; ndex < sample_size; ndex++) {
            if (sample[ndex].IsPlus()) {
                divergence += (1.0 - this->Predict(sample, ndex));
            } else {
                divergence += this->Predict(sample, ndex);
            }
        }
    }
    double loglikelyhood = 0.0;
//    double likelyhood = 1.0;
    for(std::uint32_t ndex = 0; ndex < sample_size; ndex++) {
        if (sample[ndex].IsPlus()) {
            loglikelyhood -= std::log(1.0 + std::exp(-w0 - w1*sample[ndex][pos1] - w2*sample[ndex][pos2]));
//            likelyhood *= 1.0/(1.0 + std::exp(-w0 - w1*sample[ndex][pos1] - w2*sample[ndex][pos2]));
        } else {
            loglikelyhood -= std::log(1.0 + std::exp(w0 + w1*sample[ndex][pos1] + w2*sample[ndex][pos2]));
//            likelyhood *= 1.0/(1.0 + std::exp(w0 + w1*sample[ndex][pos1] + w2*sample[ndex][pos2]));
        }
    }
    return loglikelyhood;
//    return likelyhood;
}

void vkf::Join::ComputeSave(    const std::string &complex_table_name,
                                const std::string &database,
                                const std::string &db_host = DEFAULT_HOST,
                                const std::string &db_user = DEFAULT_USER,
                                const std::string &db_pswd = DEFAULT_PASS
                            ) {
    std::uint32_t attrs_total = this->GetAttrTotal();
    // set up the account
    mariadb::account_ref acc = mariadb::account::create(db_host, db_user, db_pswd);
    // create connection
    mariadb::connection_ref con = mariadb::connection::create(acc);
    con->execute("USE " + database);
    // create table complex
    con->execute("DROP TABLE IF EXISTS " + complex_table_name);
    std::string sql_query("CREATE TABLE IF NOT EXISTS " + complex_table_name);
    sql_query.append(" (row_id int(8) NOT NULL AUTO_INCREMENT, ");
    sql_query.append("bias float(53) NOT NULL DEFAULT 0E0, ");
    sql_query.append("first_name varchar(255) NOT NULL DEFAULT 'null', ");
    sql_query.append("first_coeff float(53) NOT NULL DEFAULT 0E0, ");
    sql_query.append("second_name varchar(255) NOT NULL DEFAULT 'null', ");
    sql_query.append("second_coeff float(53) NOT NULL DEFAULT 0E0, ");
    sql_query.append("PRIMARY KEY(row_id)) ");
    sql_query.append("ENGINE = InnoDB");
    con->execute(sql_query);
    mariadb::statement_ref stmt = con->create_statement("INSERT INTO " + complex_table_name + " VALUES (NULL, ?, ?, ?, ?, ?)");
    for (std::size_t ndex = 1; ndex < attrs_total; ndex++) {
        vkf::SemiRegress base1_attr(ndex);
        double base1_likelyhood = base1_attr.Optimize(*this, 0.01);
        for (std::size_t indx = 0; indx < ndex; indx++) {
            vkf::SemiRegress base2_attr(indx);
            double base2_likelyhood = base2_attr.Optimize(*this, 0.01);
            vkf::Regress new_attr(indx, ndex);
            double likelyhood = new_attr.Optimize(*this, 0.01);
//            if (1-std::exp(2.0*(base1_likelyhood-likelyhood)/this->size()) > 0.01 && 
//                1-std::exp(2.0*(base2_likelyhood-likelyhood)/this->size()) > 0.01) /* Cox & Snell R-squared */ {
            if (1-likelyhood/base1_likelyhood > 0.01 && 1-likelyhood/base2_likelyhood > 0.01) /* McFadden's R-squared */ {
//                std::cout << "w0: " << base1_attr.GetBias() << ", w1: " << base1_attr.GetCoeff() << std::endl;
//                std::cout << "base1_likelyhood: " << base1_likelyhood << std::endl;
//                std::cout << "w0: " << base2_attr.GetBias() << ", w1: " << base2_attr.GetCoeff() << std::endl;
//                std::cout << "base2_likelyhood: " << base2_likelyhood << std::endl;
//                std::cout << "w0: " << new_attr.GetBias() << ", w1: " << new_attr.GetFirstCoeff() << ", w2: " << new_attr.GetSecondCoeff() << std::endl;
//                std::cout << "likelyhood: " << likelyhood << std::endl;
//                std::cout << "1st attribute: " << this->GetAttrName(indx) << ", 2nd attribute: " << this->GetAttrName(ndex) << std::endl;
                stmt->set_double(0, new_attr.GetBias());
                stmt->set_string(1, this->GetAttrName(indx));
                stmt->set_double(2, new_attr.GetFirstCoeff());
                stmt->set_string(3, this->GetAttrName(ndex));
                stmt->set_double(4, new_attr.GetSecondCoeff());
                stmt->insert();
            }
        }
    }
//    stmt->~statement();
    con->~connection();
    acc->~account();
    return;
};
